#SketchyFFD
#Copyright 2010-2013 Chris Phillips

#Update 4 Feb 2010 by Glenn Babcock
#change log
# - fixed issues with control point groups and dicegroups sometimes being out of position

#Update 2 Feb 2010 by Glenn Babcock
#change log
#  - added feature that combines move and transform into one operation so it can be undone all at once
#  - disabled screen refreshes during heavy operations to improve performance
#  - reorganized code
#  - began to rationalize variables
#  - added $debugFFB global

#Update 1 Feb 2010 by Glenn Babcock
#change log
#  - fixed bug where control group sizes could not be changed or where all dimensions had to be equal

### 20110810...
### TIG tweaked to module and observers sorted and $ > @@ etc...
### 20130203...
### TIG NxN issues with NaN & Infinity fixed, when selected group is 2d & height=1
### NxN now remembers last used values for w/d/h/subdivide...

require 'sketchup.rb'

module SketchyFFD ###
 
if Sketchup.version.to_f < 7.0
  UI.messagebox("SketchyFFD needs Sketchup >=7.0.\nVisit sketchup.google.com and upgrade.")
  #return nil
end

@@debugFFD=false ### make true to get lots of messages!

# Show the Ruby Console at startup for debugFFDging
Sketchup.send_action("showRubyPanel:") if @@debugFFD

#Infinity = 1.0/0.0 ### NOT needed

# VARIABLE DEFINITIONS ### some set as @@
# grp = selected group to be deformed
 @@ffdGroup=nil ### used to access the group to be deformed
# numControlPoints = number of control points in x, y, z axis
# numControlSections = numControlPoints - 1 in each dimension
# bSubdivide = logical to determine if user wants to subdivide selected group
# size = size of group to be deformed
 @@latticeGroup=nil ###FFD Control Point lattice
 @@allVerts=[]
 @@allVertWeights=[]
 @@observer=nil
 
 @@w=4
 @@d=4
 @@h=4
 @@s="false"
 
#initialize the program

def self.initiate()
  #Right click UI menu...
 unless file_loaded?(__FILE__)###
  UI.add_context_menu_handler{|menu|
    puts("正在创建菜单...") if @@debugFFD
    menu=menu.add_submenu("FFD自由变形...")
    if !Sketchup.active_model.selection.empty? && Sketchup.active_model.selection[0].is_a?(Sketchup::Group)
      menu.add_item("2x2 FFD"){self.startFFD(Sketchup.active_model.selection[0],[2,2,2],"false")}
      menu.add_item("3x3 FFD"){self.startFFD(Sketchup.active_model.selection[0],[3,3,3],"false")}
      menu.add_item("NxN FFD"){
        prompts = ["宽度: ","深度: ","高度: ","细分: "]
        values = [@@w, @@d, @@h, @@s]
        results = inputbox(prompts, values,["","","","true|false"], "FFD 规格")
        if results
		  @@w,@@d,@@h,@@s=results
		  @@w=1 if @@w<1
		  @@d=1 if @@d<1
		  @@h=1 if @@h<1
          self.startFFD(Sketchup.active_model.selection[0],[@@w,@@d,@@h],@@s) 
        end
      }
    end
    menu.add_item("锁定边线"){
      Sketchup.active_model.start_operation("锁定边线")
      Sketchup.active_model.selection.each{|ent|
        if ent.is_a?(Sketchup::Edge)
          ent.vertices.each{|v| v.set_attribute("SFFD","已锁定",true)}
        end
      }
      Sketchup.active_model.commit_operation
    }    
    menu.add_item("解锁边线"){
      Sketchup.active_model.start_operation("解锁边线")
      Sketchup.active_model.selection.each{|ent|
        if ent.is_a?(Sketchup::Edge)
          ent.vertices.each{|v| v.set_attribute("SFFD","已锁定",false)}
        end
      }
      Sketchup.active_model.commit_operation
    }   
    menu.add_item("创建路径.."){
      prompts = ["宽度","深度","单元格宽度","单元格深度"]
      values = [4, 4, 10, 10]
      results = UI.inputbox(prompts, values, " FFD 路径规格")
      Sketchup.active_model.start_operation("创建 FFD 路径")
      grp=Sketchup.active_model.active_entities.add_group() ###
      0.upto(results[0]-2){|w|
        0.upto(results[1]-2){|h|
          xform=Geom::Transformation.new([w*results[2],h*results[3],0])
          xform=xform * Geom::Transformation.scaling(results[2],results[3], 1.0)
          f=[[0,0,0].transform(xform),
             [0,1,0].transform(xform),
             [1,1,0].transform(xform),
             [1,0,0].transform(xform) ]
          grp.entities.add_face(f)
        }
      }
      Sketchup.active_model.commit_operation
      Sketchup.active_model.selection.clear
      Sketchup.active_model.selection.add(grp)
      if (results)
        self.startFFD(Sketchup.active_model.selection[0],[results[0],results[1],1]) 
      end
    }
  }
 end#unless
 file_loaded(__FILE__)###
end

#called from context menu to create a control lattice and calculate the vertex weights for a group.
def self.startFFD(grp, numControlPoints, bSubdivide)
  
    puts("in startFFD...") if @@debugFFD
    
  #put group to be deformed into global so all methods can access it
  @@ffdGroup=grp 
  #@@ffdGroup.name="ffdGroup"
  #make sure this group is unique or chaos ensues.
  begin ###
    @@ffdGroup.make_unique if @@ffdGroup.entities.parent.instances[1]
  rescue
    ###
  end
    #calculate size of the Def group
    sizeDefG=([@@ffdGroup.bounds.width,@@ffdGroup.bounds.height,@@ffdGroup.bounds.depth])
    puts sizeDefG if @@debugFFD
  
  Sketchup.active_model.start_operation("Create FFD group",true)
  
    numControlSections=[numControlPoints[0]-1,numControlPoints[1]-1,numControlPoints[2]-1]
    
    #if user selected Subdivide, do it
    self.dice_group(sizeDefG,numControlSections) if(bSubdivide=="true")
  
    #calculate size of the Def group
    sizeDefG=([@@ffdGroup.bounds.width,@@ffdGroup.bounds.height,@@ffdGroup.bounds.depth])

      puts("after dice_group") if @@debugFFD
      puts sizeDefG if @@debugFFD

    #create CP lattice group
    @@latticeGroup=self.createControlLattice(sizeDefG,numControlSections)  
    @@latticeGroup.name="FFD control points"
    
    #store the current transformation in an attribute
    @@latticeGroup.set_attribute("controlLattice","currentTransformation",@@ffdGroup.transformation.to_a)

    puts("in startFFD after createControlLattice") if @@debugFFD
    
    #calculate the D group's vertex weights
    self.initFFD(sizeDefG,numControlSections)   
    
  Sketchup.active_model.commit_operation
  
end

  #called from startFFD to calculate the groups vertex weights.    
  def self.initFFD(sizeDefG,numControlPoints)
      puts("in initFFD...") if @@debugFFD
      
    #create arrary of vertices
    @@allVerts=[]
    @@ffdGroup.entities.each{|ent|
      if ent.is_a?(Sketchup::Edge) && ent.curve
        ent.explode_curve #all curves need to be exploded for deform to work right.
      end
    }
    @@ffdGroup.entities.each{|ent|
      begin #guard against entities that dont have verts.
        ent.vertices.each{|vert|
          @@allVerts.push(vert)
        }
      rescue
        puts "Warning:entities of type '#{ent.typename}' can't be deformed."
      end
    }
    #allverts now contains redundant verts; remove duplicates.
    @@allVerts.uniq!
    
      puts @@allVerts.length if @@debugFFD
    
    #Global to hold the calculated weight for each vertex per control point.
    @@allVertWeights=[]
    
    #this loop could be dramaticly optimized.
    vi=0        
    @@allVerts.each{|vert|
      stuv=vert.position
      stuv.x=stuv.x/sizeDefG[0];stuv.y=stuv.y/sizeDefG[1];stuv.z=stuv.z/sizeDefG[2]
      #calc weights.
      stuv.x=0.0 if(stuv.x.to_f.nan?)
      stuv.y=0.0 if(stuv.y.to_f.nan?)
      stuv.z=0.0 if(stuv.z.to_f.nan?)
      stuv.x=1.0 if(stuv.x.to_f.infinite?)
      stuv.y=1.0 if(stuv.y.to_f.infinite?)
      stuv.z=1.0 if(stuv.z.to_f.infinite?)
	  
      puts stuv.inspect if @@debugFFD
      weights=[]
      0.upto(numControlPoints[0]){|x|
        bx=self.calcBernstein(x,numControlPoints[0],stuv.x)
        0.upto(numControlPoints[1]){|y|
          bxy=bx*self.calcBernstein(y,numControlPoints[1],stuv.y)
          0.upto(numControlPoints[2]){|z|
            weights.push(bxy * self.calcBernstein(z,numControlPoints[2],stuv.z))
          }
        }
      }
      @@allVertWeights.push(weights)
      vert.set_attribute("SFFD","locked",false)
      
      vi=vi+1
      Sketchup.set_status_text("Weighing  #{vi} of #{@@allVerts.length}") if(vi%100==0)
  
      }
  end
  
    #called by initFFD to calculate the Bernstein polynomial.
    #Uses a table for speed. 
    #thanks to steven-arts for the idea.
    def self.calcBernstein(i, n, u) #Bernstein Polynomial
      binomialTable=[] #1 Feb 2010 initialize the array on each call
      begin
        binomial=binomialTable[i][n]
      rescue
        binomialTable[i]=[] if(binomialTable[i]==nil)
        binomialTable[i][n]=self.ffdfactorial(n).to_f / (self.ffdfactorial(n - i).to_f * self.ffdfactorial(i).to_f)
        binomial=binomialTable[i][n]
      end
      #binomial = factorial(n).to_f / (factorial(n - i).to_f * factorial(i).to_f)
      bernstein = binomial * (u**i) * ((1-u)**(n-i))
      return(bernstein)
    end
  
      #called by calcBernstein. Uses a table for speed. 
      #thanks to steven-arts for the idea.
      def self.ffdfactorial(n)
        ffdfactorialTable = []
        sum=ffdfactorialTable[n]
        return sum if sum
        
        sum = 1
        sum.upto(n) { |i| sum *= i }
        ffdfactorialTable[n]=sum
        return sum
      end

  #Called from startFFD, subdivides goemetry when input parameter is true
  def self.dice_group(sizeDefG,numControlSections)

      puts("in dice_group") if @@debugFFD
    
    #find the origin of the D group
    ffdGroupOrigin=@@ffdGroup.bounds.min
    diceGroup=Sketchup.active_model.active_entities.add_group()
    diceGroup=diceGroup.move!(ffdGroupOrigin)
    diceGroup.name="diceGroup"
    
    #calculate the size of the edges for the diceGroup
    xstep=sizeDefG[0].to_f/numControlSections[0]
    ystep=sizeDefG[1].to_f/numControlSections[1]    
    zstep=sizeDefG[2].to_f/numControlSections[2]
    
    #handle dim of 0 (1)
    xstep=0.0 if(xstep.to_f.nan?)
    ystep=0.0 if(ystep.to_f.nan?)
    zstep=0.0 if(zstep.to_f.nan?)
    xstep=1.0 if(xstep.to_f.infinite?)
    ystep=1.0 if(ystep.to_f.infinite?)
    zstep=1.0 if(zstep.to_f.infinite?)
   
    #create the diceGroup faces
    1.upto(numControlSections[0]-1){|x|
      diceGroup.entities.add_face([[x*xstep,0,0],[x*xstep,0,sizeDefG[2]],[x*xstep,sizeDefG[1],sizeDefG[2]],[x*xstep,sizeDefG[1],0]])
    }
    1.upto(numControlSections[1]-1){|y|
      diceGroup.entities.add_face([[0,y*ystep,0],[0,y*ystep,sizeDefG[2]],[sizeDefG[0],y*ystep,sizeDefG[2]],[sizeDefG[0],y*ystep,0]])
    }
    1.upto(numControlSections[2]-1){|z|
      diceGroup.entities.add_face([[0,0,z*zstep],[0,sizeDefG[1],z*zstep],[sizeDefG[0],sizeDefG[1],z*zstep],[sizeDefG[0],0,z*zstep]])
    }
    
      puts("@@ffdGroup.transformation:") if @@debugFFD
      puts @@ffdGroup.transformation.to_a if @@debugFFD
      
    #intersect the diceGroup with the D group
    @@ffdGroup.entities.intersect_with(true,@@ffdGroup.transformation,@@ffdGroup.entities,@@ffdGroup.transformation,false,[diceGroup])
    diceGroup.erase!

  end


  #Called by startFFD create a group to act as a control lattice.
  def self.createControlLattice(sizeDefG, numControlSections)
    #remove previous lattice (if any)
    if(@@latticeGroup)
      begin
        @@latticeGroup.erase!
        @@ffdGroup.remove_observer(@@observer)
      rescue
        #group was already deleted.
      end
    end

      puts("in createControlLattice") if @@debugFFD     
      puts("numControlSections:") if @@debugFFD
      puts numControlSections if @@debugFFD
    
    latticeGroup=Sketchup.active_model.active_entities.add_group()
    
    #create control lattice group.
    entities=latticeGroup.entities
    
    #create an @@observer for events. Not used yet.
    @@observer=self::LatticeObserver.new(@@ffdGroup, latticeGroup)
    
    @@ffdGroup.add_observer(@@observer)
    
    xstep=sizeDefG[0].to_f/numControlSections[0]
    ystep=sizeDefG[1].to_f/numControlSections[1]    
    zstep=sizeDefG[2].to_f/numControlSections[2]
    
    #handle dim of 0 (1)
	xstep=0.0 if(xstep.to_f.nan?)
    ystep=0.0 if(ystep.to_f.nan?)
    zstep=0.0 if(zstep.to_f.nan?)
    xstep=1.0 if(xstep.to_f.infinite?)
    ystep=1.0 if(ystep.to_f.infinite?)
    zstep=1.0 if(zstep.to_f.infinite?)
    
    #get the location of origin of the fffGroup
    ffdGroupOrigin=[]
    ffdGroupOrigin=@@ffdGroup.bounds.min.to_a
    
    controlPoints=[]
    0.upto(numControlSections[0]){|x|
      0.upto(numControlSections[1]){|y|
        0.upto(numControlSections[2]){|z|
          
          #NOTE: 0.001+ is because bug in sketchup will not allow construction point at 0,0,0
          controlPoints.push([ffdGroupOrigin[0]+0.001+(x*xstep),ffdGroupOrigin[1]+0.001+(y*ystep),ffdGroupOrigin[2]+0.001+(z*zstep)])
        }
      }
    }
    

    #create construction points.
    index=0    
    controlPoints.each{|lcp|
      cpt=entities.add_cpoint(Geom::Point3d.new([lcp[0],lcp[1],lcp[2]]))
      cpt.set_attribute("controlPoint","originalPosition",cpt.position.to_a)
      cpt.set_attribute("controlPoint","index",index)
      index=index+1
      #trigger event any time the user moves a point.
      cpt.add_observer(@@observer) 
    }
    
      puts("before moving latticeGroup to match ffdGroup") if @@debugFFD
  
    #move lattice group to match ffd group    
    #t = @@ffdGroup.bounds.min
    #latticeGroup=latticeGroup.move!(t)
    
    latticeGroup.set_attribute("controlLattice","currentTransformation",latticeGroup.transformation.to_a)
    
    #monitor group open/close event. not used yet.
    latticeGroup.add_observer(@@observer)

    return(latticeGroup)
  end

### OBSERVER FIX!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
#this class provides events when the user manipulates the control points
class self::LatticeObserver < Sketchup::EntityObserver ###
  ###
  def initialize(ffdGroup=nil, latticeGroup=nil) ### pass groups to observer!
    @ffdGroup=ffdGroup if ffdGroup
    @latticeGroup=latticeGroup if latticeGroup
  end
  ### ENTITIESobserver methods ONLY !
  #def onElementAdded(entity, xx)
  #end
  #def onContentsModified (entity)
  #end
  #called when lattice point is moved.
  def onChangeEntity(entity)
    return nil if not @latticeGroup or not @latticeGroup.valid?
    ###puts("in onChangeEntity") if @@debugFFD
    ###puts "changed "+entity.to_s if @@debugFFD
    if entity==@latticeGroup
      #entity.class==Sketchup::ComponentDefinition)
      ###puts("in onChangeEntity and entity.class==Sketchup::ComponentDefinition") if @@debugFFD
      ###puts entity.name if @@debugFFD
      entity.set_attribute("controlLattice","currentTransformation",entity.transformation.to_a)   
    elsif entity.is_a?(Sketchup::ConstructionPoint) and entity.parent==@latticeGroup.entities.parent
      ###
      begin
        SketchyFFD.updateFFD() if @ffdGroup and @ffdGroup.valid?###
      rescue
        ###
      end
      ###
    end
  end
  def onEraseEntity(entity)
    begin
      @latticeGroup.erase! if @latticeGroup and @latticeGroup.valid? ###
    rescue
      ###
    end
  end
  ### INSTANCEobserver methods oNLY !
  #called when lattice group is opened
  #def onOpen(instance)
    #instance.set_attribute("controlLattice","isOpen",true)
  #end
  #called when lattice group is closed
  #def onClose(instance)
    #instance.set_attribute("controlLattice","isOpen",false)
  #end
end   
### !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!


  #called from onChangeEntity event in LatticeObserver class to deform the group 
  #based on the position changes of the control points.
  def self.updateFFD()
    ### trap for erased group etc...
    if not @@ffdGroup or not @@ffdGroup.valid?
      return nil
    end
    #disable UI, combine this operation with previous for Undo
    Sketchup.active_model.start_operation("Update FFD",true,false,true)
    
    puts("in updateFFD") if @@debugFFD
    puts("@@latticeGroup.tranformation, @@ffdGroup.transformation") if @@debugFFD
    puts @@latticeGroup.transformation if @@debugFFD
    puts @@ffdGroup.transformation if @@debugFFD

    xform=Geom::Transformation.new()
    ### disabled originally ??
    #if(@@latticeGroup.get_attribute("controlLattice","isOpen",false))
      #if group is open transform by inverse of saved xform (the real xform).
      
        #puts("@@latticeGroup isOpen, transform inverse of saved xform") if @@debugFFD
      
      #xform=Geom::Transformation.new(@@latticeGroup.get_attribute("controlLattice","currentTransformation",nil))
      #xform.invert!
    #end
    
    deltas=[]
    @@latticeGroup.entities[0].parent.entities.each{|cpt|
      op=cpt.get_attribute("controlPoint","originalPosition",nil)
      
        puts("op:") if @@debugFFD
        puts op if @@debugFFD
        
      if(op)#ent is a control point.
        #cp=cpt.position.transform(xform).to_a#get current position
        cp=cpt.position.to_a
        
          puts("cp, cp2:") if @@debugFFD
          puts cp if @@debugFFD
          
        if(cp!=op)#point moved?
          delta=[]#calculate change
          delta[0]=cp[0]-op[0]
          delta[1]=cp[1]-op[1]
          delta[2]=cp[2]-op[2]
          #update stored position with new position
          cpt.set_attribute("controlPoint","originalPosition",cp)
          index=cpt.get_attribute("controlPoint","index",nil)
          deltas.push([index,delta])#add to array of changes.
        end
      end
    }
    puts("deltas.inspect:") if @@debugFFD
    puts deltas.inspect if @@debugFFD
    #update mesh    
    self.applyMultipleFFD(deltas) if (deltas.length>0)
      
    Sketchup.active_model.commit_operation
  end
  
    #called from updateFFD to deform the mesh based on the change in position of each control point.
    #deltas is an array. 
    #each element in the deltas array is an array with 2 elements. 
    #the first element is the index of the control point.
    #the second element is the change in postition of that point. (a array of 3 floats).
    #example:
    #deltas=[]
    #deltas.push([0,[0.0,0.0,1.0]])  {move control point 0 "up" 1.0}
    def self.applyMultipleFFD(deltas)
      #grp.entities.transform_entities(Geom::Transformation.new([0.0,0.0,0.001]),grp.entities.to_a)
      vi=0;#used to loop through @@allVertWeights in the same order as @@allVerts
      @@allVerts.each{|vert|
        weights=@@allVertWeights[vi] #get pre-calculated array of vertex weights (one per control point)
        vi=vi+1
        next if(vert.get_attribute("SFFD","locked",false))
        #accumulate the change in position of this vertex based on 
        #each control points weighted movement.
        dvect=[0.0,0.0,0.0] 
        deltas.each{|delta|
          weight=weights[delta[0]] #delta[0] = control point index.
          #puts weight            
          vect=delta[1]#delta[1]=delta vector for control point
          dvect[0]+=vect[0]*weight
          dvect[1]+=vect[1]*weight
          dvect[2]+=vect[2]*weight
        }
        #move the vertex.
        begin
          @@ffdGroup.entities.transform_entities(Geom::Transformation.new(dvect),vert)
        rescue
         ###
        end
        #display progress
        Sketchup.set_status_text("Deforming #{vi} of #{@@allVerts.length}") if(vi%100==0)
        #puts("Deforming #{vi} of #{@@allVerts.length}") if(vi%100==0)
      }

  end
  
self.initiate()

end#module
#end
